from __future__ import absolute_import

import wrapt

from thundra import utils, constants
from thundra.integrations.mysql import MysqlIntegration

mysql_integration = MysqlIntegration()

class MysqlCursorWrapper(wrapt.ObjectProxy):

    def __init__(self, cursor, connection_wrapper):
        super(MysqlCursorWrapper, self).__init__(cursor)
        self._self_connection = connection_wrapper

    def execute(self, *args, **kwargs):
        response = mysql_integration.create_span(
            self.__wrapped__.execute,
            self._self_connection,
            args,
            kwargs,
        )
        return response


    def __enter__(self):
        # raise appropriate error if api not supported (should reach the user)
        self.__wrapped__.__enter__  # pylint: disable=W0104
        return self

class MysqlConnectionWrapper(wrapt.ObjectProxy):
    def cursor(self, *args, **kwargs):
        cursor = self.__wrapped__.cursor(*args, **kwargs)
        return MysqlCursorWrapper(cursor, self)

def _wrapper(wrapped, instance, args, kwargs):
    connection = wrapped(*args, **kwargs)
    return MysqlConnectionWrapper(connection)


def patch():
    disable_rdb_integration_by_env = utils.get_configuration(constants.THUNDRA_DISABLE_RDB_INTEGRATION)
    if not utils.should_disable(disable_rdb_integration_by_env):
        wrapt.wrap_function_wrapper(
            'mysql.connector',
            'connect',
            _wrapper)


from __future__ import absolute_import
from importlib import import_module
import thundra.integrations.modules as integrations
import thundra.listeners as listeners


def _import_exists(module_name):
    """
    Validates if import module exists
    :param module_name: module name to import
    :return: Bool
    """
    try:
        import_module(module_name)
        return True
    except ImportError:
        return False


for patch_module in integrations.MODULES:
    if _import_exists(patch_module):
        integrations.MODULES[patch_module].patch()



import os
import sys
import logging
from urllib.parse import urlparse
from thundra import constants

logger = logging.getLogger(__name__)

def get_configuration(key, default=None):
    return os.environ.get(key, default=default)


def should_disable(disable_by_env, disable_by_param=False):
    if disable_by_env == 'true':
        return True
    elif disable_by_env == 'false':
        return False
    return disable_by_param


def get_application_id(context):
    aws_lambda_log_stream_name = getattr(context, constants.CONTEXT_LOG_STREAM_NAME, None)
    application_id = aws_lambda_log_stream_name.split("]")[1] if aws_lambda_log_stream_name is not None else ''
    return application_id


def get_aws_lambda_function_memory_size():
    return os.environ.get(constants.AWS_LAMBDA_FUNCTION_MEMORY_SIZE)


#### memory ####
def process_memory_usage():
    try:
        with open('/proc/self/status', 'r') as procfile:
            mems = {}
            for line in procfile:
                fields = line.split(':')
                try:
                    mem_key = fields[0]
                    mem_val = (fields[1].split())[0]
                    mems[mem_key] = mem_val
                except IndexError:
                    continue
            
            size_from_env_var = get_aws_lambda_function_memory_size()
            if not size_from_env_var:
                size = int(mems['VmSize'])
                size_in_bytes = int(size * 1024)
            else:
                size_in_bytes = int(float(size_from_env_var) * 1048576)

            used_mem_in_kb = int(mems['VmRSS']) + int(mems['VmSwap'])
            used_mem_in_bytes = used_mem_in_kb * 1024

            return size_in_bytes, used_mem_in_bytes
    except IOError as e:
        logger.error('ERROR: {}', e)
        return 0, 0

##### cpu #####
def process_cpu_usage():
    try:
        with open('/proc/self/stat', 'r') as procfile:
            process_cpu_usages = procfile.readline()
            # get utime from /proc/<pid>/stat, 14 item
            u_time = process_cpu_usages.split(' ')[13]
            # get stime from proc/<pid>/stat, 15 item
            s_time = process_cpu_usages.split(' ')[14]
            # count total process used time
            process_cpu_used = int(u_time) + int(s_time)
            return (float(process_cpu_used))
    except IOError as e:
        logger.error('ERROR: {}', e)
        return 0


def system_cpu_usage():
    try:
        with open('/proc/stat', 'r') as procfile:
            system_cpu_usages = procfile.readline()
            system_cpu_used = 0
            system_cpu_total = 0
            count = 0
            for usage in system_cpu_usages.split(' ')[2:]:
                if count == 5:
                    break
                elif count != 3 and count != 4:
                    system_cpu_used += int(usage)
                system_cpu_total += int(usage)
                count += 1
            return float(system_cpu_total), float(system_cpu_used)
    except IOError as e:
        logger.error('ERROR: {}', e)
        return 0, 0


#####################################################################
###
#####################################################################


class Singleton(object):
    _instances = {}
    def __new__(class_, *args, **kwargs):
        if class_ not in class_._instances:
            class_._instances[class_] = super(Singleton, class_).__new__(class_, *args, **kwargs)
        return class_._instances[class_]


def get_all_env_variables():
    return os.environ


def get_module_name(module):
    return module.__name__


def string_to_list(target, indicator):
    return target.split(indicator)


def str2bool(val):
    if val.lower() in ("yes", "true", "t", "1"):
        return True
    elif val.lower() in ("no", "false", "f", "0"):
        return False
    raise ValueError


def process_trace_def_env_var(value):
    value = value.strip().split('[')
    path = value[0].split('.')
    trace_args = {}

    function_prefix = path[-1][:-1] if path[-1] != '*' else ''
    module_path = ".".join(path[:-1])
    trace_string = value[1].strip(']').split(',')
    for arg  in trace_string:
        arg = arg.split('=')
        try:
            trace_args[arg[0]] = arg[1]
        except:
            pass

    return module_path, function_prefix, trace_args


def get_allowed_functions(module):
    allowed_functions = []
    for prop in vars(module):
        #TO DO: differentiate functions
        allowed_functions.append(str(prop))
    return allowed_functions


def get_aws_region_from_arn(func_arn):
    return func_arn.split(':')[3] if len(func_arn.split(':')) >= 3 else ""

def is_excluded_url(url):
    host = urlparse(url).netloc
    for method in EXCLUDED_URLS:
        for excluded_url in EXCLUDED_URLS[method]:
            if method(host, excluded_url):
                return True
    return False

# Excluded url's 
EXCLUDED_URLS = {
    str.endswith: [
        'thundra.io',
    ],
    str.__contains__: [
        'amazonaws.com',
        'accounts.google.com',
    ],
}



import time
from threading import Lock

import opentracing
from thundra.opentracing.recorder import RecordEvents


class ThundraSpan(opentracing.Span):

    def __init__(self,
                 tracer,
                 operation_name=None,
                 class_name=None,
                 domain_name=None,
                 context=None,
                 tags=None,
                 start_time=None,
                 span_order=-1):
        super(ThundraSpan, self).__init__(tracer, context)
        self._tracer = tracer
        self._context = context
        self._lock = Lock()
        self.operation_name = operation_name
        self.class_name = class_name
        self.domain_name = domain_name
        self.start_time = start_time or int(time.time() * 1000)
        self.finish_time = 0
        self.span_order = span_order
        self.tags = tags if tags is not None else {}
        self.logs = []

    @property
    def context(self):
        return self._context

    @property
    def trace_id(self):
        return self._context.trace_id

    @property
    def transaction_id(self):
        return self._context.transaction_id

    @property
    def span_id(self):
        return self._context.span_id

    def set_operation_name(self, operation_name):
        with self._lock:
            self.operation_name = operation_name
        return super(ThundraSpan, self).set_operation_name(operation_name)

    def set_tag(self, key, value):
        with self._lock:
            if self.tags is None:
                self.tags = {}
            self.tags[key] = value
        return super(ThundraSpan, self).set_tag(key, value)

    def get_tag(self, key):
        return self.tags.get(key)

    def finish(self, f_time=None):
        with self._lock:
            self.finish_time = int(time.time() * 1000) if f_time is None else f_time
            self._tracer.record(RecordEvents.FINISH_SPAN, self)

    def log_kv(self, key_values, timestamp=None):
        with self._lock:
            log = key_values
            log['timestamp'] = timestamp
            self.logs.append(log)
        return super(ThundraSpan, self).log_kv(key_values, timestamp)

    def set_baggage_item(self, key, value):
        new_context = self.context.context_with_baggage_item(key, value)
        with self._lock:
            self._context = new_context
        return self

    def get_baggage_item(self, key):
        with self._lock:
            return self.context.baggage.get(key)

    def set_error_to_tag(self, err):
        error_type = type(err)
        self.set_tag('error', True)
        self.set_tag('error.kind', error_type.__name__)
        self.set_tag('error.message', str(err))

    def get_duration(self):
        if self.finish_time == 0:
            return int(time.time() * 1000) - self.start_time
        else:
            return self.finish_time - self.start_time


_tags = {}

def set_tag(key, value):
    _tags[key] = value

def set_many(tags):
    _tags.update(tags)

def get_tag(key):
    if key in _tags:
        return _tags[key]
    return None

def get_tags_dict():
    return _tags.copy()

def remove_tag(key):
    return _tags.pop(key, None)

def clear():
    _tags.clear()


import time
import uuid
import sys

import json
from thundra import constants, utils
import thundra.plugins.invocation.invocation_support as invocation_support
import thundra.application_support as application_support


class InvocationPlugin:

    def __init__(self):
        self.hooks = {
            'before:invocation': self.before_invocation,
            'after:invocation': self.after_invocation
        }
        self.start_time = 0
        self.end_time = 0
        self.invocation_data = {}

    def before_invocation(self, plugin_context):
        context = plugin_context['context']
        function_name = getattr(context, constants.CONTEXT_FUNCTION_NAME, None)
        invocation_support.clear()
        self.start_time = int(time.time() * 1000)

        self.invocation_data = {
            'id': str(uuid.uuid4()),
            'type': "Invocation",
            'agentVersion': constants.THUNDRA_AGENT_VERSION,
            'dataModelVersion': constants.DATA_FORMAT_VERSION,
            'applicationId': utils.get_application_id(context),
            'applicationDomainName': constants.AWS_LAMBDA_APPLICATION_DOMAIN_NAME,
            'applicationClassName': constants.AWS_LAMBDA_APPLICATION_CLASS_NAME,
            'applicationName': function_name,
            'applicationVersion': getattr(context, constants.CONTEXT_FUNCTION_VERSION, None),
            'applicationStage': utils.get_configuration(constants.THUNDRA_APPLICATION_STAGE, ''),
            'applicationRuntime': 'python',
            'applicationRuntimeVersion': str(sys.version_info[0]),
            'applicationTags': {},

            'traceId': plugin_context.get('trace_id', ""),
            'transactionId': plugin_context.get('transaction_id', context.aws_request_id),
            'spanId': plugin_context.get('span_id', ""),
            'functionPlatform': constants.CONTEXT_FUNCTION_PLATFORM,
            'functionName': getattr(context, 'function_name', None),
            'functionRegion': utils.get_configuration(constants.AWS_REGION, default=''),
            'duration': None, 
            'startTimestamp': int(self.start_time),
            'finishTimestamp': None,
            'erroneous': False,
            'errorType': '',
            'errorMessage': '',
            'errorCode': -1,
            'coldStart': constants.REQUEST_COUNT == 1,
            'timeout': False,
            'tags': {},
        }

    def after_invocation(self, plugin_context):
        total_mem, used_mem = utils.process_memory_usage()
        used_mem_in_mb = used_mem / 1048576
        self.end_time = time.time() * 1000
        context = plugin_context['context']
        #### ADDING USER TAGS ####
        self.invocation_data['tags'] = invocation_support.get_tags_dict()

        #### ERROR ####
        if 'error' in plugin_context:
            error = plugin_context['error']
            error_type = type(error)
            self.invocation_data['erroneous'] = True
            self.invocation_data['errorType'] = error_type.__name__
            self.invocation_data['errorMessage'] = str(error)
            if hasattr(error, 'code'):
                self.invocation_data['errorCode'] = error.code

            # Adding tags
            self.invocation_data['tags']['error'] = True
            self.invocation_data['tags']['error.kind'] = error_type.__name__
            self.invocation_data['tags']['error.message'] = str(error)
            if hasattr(error, 'code'):
                self.invocation_data['tags']['error.code'] = error.code
            if hasattr(error, 'object'):
                self.invocation_data['tags']['error.object'] = error.object
            if hasattr(error, 'stack'):
                self.invocation_data['tags']['error.stack'] = error.stack

        self.invocation_data['timeout'] = plugin_context.get('timeout', False)

        duration = self.end_time - self.start_time
        self.invocation_data['duration'] = int(duration)
        self.invocation_data['finishTimestamp'] = int(self.end_time)

        #### ADDING AWS TAGS ####
        self.invocation_data['applicationTags'] = application_support.get_application_tags()
        self.invocation_data['tags']['aws.region'] = utils.get_aws_region_from_arn(getattr (context, constants.CONTEXT_INVOKED_FUNCTION_ARN, None))
        self.invocation_data['tags']['aws.lambda.name'] = getattr(context, constants.CONTEXT_FUNCTION_NAME, None)
        self.invocation_data['tags']['aws.lambda.arn'] = getattr(context, constants.CONTEXT_INVOKED_FUNCTION_ARN, None)
        self.invocation_data['tags']['aws.lambda.memory_limit'] = getattr(context, constants.CONTEXT_MEMORY_LIMIT_IN_MB, None)
        self.invocation_data['tags']['aws.lambda.log_group_name'] = getattr(context, constants.CONTEXT_LOG_GROUP_NAME, None)
        self.invocation_data['tags']['aws.lambda.log_stream_name'] = getattr(context, constants.CONTEXT_LOG_STREAM_NAME, None)
        self.invocation_data['tags']['aws.lambda.invocation.cold_start'] = self.invocation_data['coldStart']
        self.invocation_data['tags']['aws.lambda.invocation.timeout'] = plugin_context.get('timeout', False)
        self.invocation_data['tags']['aws.lambda.invocation.request_id'] = getattr(context, constants.CONTEXT_AWS_REQUEST_ID, None)
        self.invocation_data['tags']['aws.lambda.invocation.memory_usage'] = used_mem_in_mb

        reporter = plugin_context['reporter']
        report_data = {
            'apiKey': reporter.api_key,
            'type': 'Invocation',
            'dataModelVersion': constants.DATA_FORMAT_VERSION,
            'data': self.invocation_data
        }
        reporter.add_report(json.loads(json.dumps(report_data)))
        invocation_support.clear()



from __future__ import absolute_import
import wrapt
from thundra import utils
from thundra import constants
from thundra.integrations.redis import RedisIntegration

redis_integration = RedisIntegration()
def _wrapper(wrapped, instance, args, kwargs):
    response = redis_integration.create_span(
        wrapped,
        instance,
        args,
        kwargs
    )
    return response


def patch():
    disable_redis_integration_by_env = utils.get_configuration(constants.THUNDRA_DISABLE_REDIS_INTEGRATION)
    if not utils.should_disable(disable_redis_integration_by_env):
        for method in map(str.lower, constants.RedisCommandTypes.keys()):
            try:
                wrapt.wrap_function_wrapper(
                    'redis.client',
                    'Redis.' + method,
                    _wrapper
                )
            except:
                pass



from __future__ import absolute_import
import wrapt
from thundra import utils
from thundra import constants
from thundra.integrations.requests import RequestsIntegration

request_integration = RequestsIntegration()


def _wrapper(wrapped, instance, args, kwargs):
    prepared_request = args[0]

    if utils.is_excluded_url(prepared_request.url):
        return wrapped(*args, **kwargs)

    response = request_integration.create_span(
        wrapped,
        instance,
        args,
        kwargs,
    )
    return response

def patch():
    disable_http_integration_by_env = utils.get_configuration(constants.THUNDRA_DISABLE_HTTP_INTEGRATION)
    if not utils.should_disable(disable_http_integration_by_env):
        wrapt.wrap_function_wrapper(
            'requests',
            'Session.send',
            _wrapper
        )
        


import uuid
import threading
import gc
import time
import sys

from thundra import utils, constants
from thundra.opentracing.tracer import ThundraTracer
import thundra.application_support as application_support


class MetricPlugin:

    def __init__(self):
        self.hooks = {
            'before:invocation': self.before_invocation,
            'after:invocation': self.after_invocation
        }
        self.system_cpu_usage_start = float(0)
        self.system_cpu_usage_end = float(0)
        self.system_cpu_total_start = float(0)
        self.system_cpu_total_end = float(0)
        self.process_cpu_usage_start = float(0)
        self.process_cpu_usage_end = float(0)
        self.metric_data = {}
        self.tracer = ThundraTracer.get_instance()

    def before_invocation(self, plugin_context):
        context = plugin_context['context']
        function_name = getattr(context, constants.CONTEXT_FUNCTION_NAME, None)
        metric_time = time.time() * 1000

        active_span = self.tracer.get_active_span()

        self.metric_data = {
            'type': "Metric",
            'agentVersion': constants.THUNDRA_AGENT_VERSION,
            'dataModelVersion': constants.DATA_FORMAT_VERSION,
            'applicationId': utils.get_application_id(context),
            'applicationDomainName': constants.AWS_LAMBDA_APPLICATION_DOMAIN_NAME,
            'applicationClassName': constants.AWS_LAMBDA_APPLICATION_CLASS_NAME,
            'applicationName': function_name,
            'applicationVersion': getattr(context, constants.CONTEXT_FUNCTION_VERSION, None),
            'applicationStage': utils.get_configuration(constants.THUNDRA_APPLICATION_STAGE, ''),
            'applicationRuntime': 'python',
            'applicationRuntimeVersion': str(sys.version_info[0]),
            'applicationTags': {},

            'traceId': active_span.trace_id if active_span is not None else '',
            'transactionId': plugin_context.get('transaction_id', context.aws_request_id),
            'spanId': active_span.span_id if active_span is not None else '',
            'metricTimestamp': int(metric_time),
            'tags': {
                'aws.region': utils.get_configuration(constants.AWS_REGION, default='')
            }
        }
        self.system_cpu_total_start, self.system_cpu_usage_start = utils.system_cpu_usage()
        self.process_cpu_usage_start = utils.process_cpu_usage()

    def after_invocation(self, plugin_context):
        self.metric_data['applicationTags'] = application_support.get_application_tags()
        reporter = plugin_context['reporter']
        self.add_thread_metric_report(reporter)
        self.add_gc_metric_report(reporter)
        self.add_memory_metric_report(reporter)
        self.add_cpu_metric_report(reporter)

    def add_thread_metric_report(self, reporter):
        active_thread_counts = threading.active_count()
        thread_metric_data = {
            'id': str(uuid.uuid4()),
            'metricName': 'ThreadMetric',
        }
        metrics = {
            'threadCount': active_thread_counts if active_thread_counts is not None else -1
        }
        thread_metric_data.update(self.metric_data)
        thread_metric_data['metrics'] = metrics
        thread_metric_report = {
            'data': thread_metric_data,
            'type': 'Metric',
            'apiKey': reporter.api_key,
            'dataModelVersion': constants.DATA_FORMAT_VERSION
        }
        reporter.add_report(thread_metric_report)

    def add_gc_metric_report(self, reporter):
        gc_metrics = gc.get_stats()
        gc_metric_data = {
            'id': str(uuid.uuid4()),
            'metricName': 'GCMetric'
        }
        gen = 0
        metrics = {}
        for metric in gc_metrics:
            key = 'generation' + str(gen) + 'Collections'
            metrics[key] = metric['collections']
            gen += 1
        gc_metric_data.update(self.metric_data)
        gc_metric_data['metrics'] = metrics
        gc_metric_report = {
            'data': gc_metric_data,
            'type': 'Metric',
            'apiKey': reporter.api_key,
            'dataModelVersion': constants.DATA_FORMAT_VERSION
        }
        reporter.add_report(gc_metric_report)

    def add_memory_metric_report(self, reporter):
        size, used = utils.process_memory_usage()
        memory_metric_data = {
            'id': str(uuid.uuid4()),
            'metricName': 'MemoryMetric',
        }
        metrics = {
            'app.maxMemory': size,
            'app.usedMemory': used
        }
        memory_metric_data.update(self.metric_data)
        memory_metric_data['metrics'] = metrics
        memory_metric_report = {
            'data': memory_metric_data,
            'type': 'Metric',
            'apiKey': reporter.api_key,
            'dataModelVersion': constants.DATA_FORMAT_VERSION
        }
        reporter.add_report(memory_metric_report)

    def add_cpu_metric_report(self, reporter):
        self.process_cpu_usage_end = utils.process_cpu_usage()
        self.system_cpu_total_end, self.system_cpu_usage_end = utils.system_cpu_usage()
        process_cpu_load = 0
        system_cpu_load = 0
        system_cpu_total = self.system_cpu_total_end - self.system_cpu_total_start
        system_cpu_usage = self.system_cpu_usage_end - self.system_cpu_usage_start
        process_cpu_usage = self.process_cpu_usage_end - self.process_cpu_usage_start
        if system_cpu_total != 0:
            cpu_load = process_cpu_usage/system_cpu_total
            process_cpu_load = 1 if cpu_load > 1 else cpu_load
            cpu_load = system_cpu_usage/system_cpu_total
            system_cpu_load = 1 if cpu_load > 1 else cpu_load

        cpu_metric_data = {
            'id': str(uuid.uuid4()),
            'metricName': 'CPUMetric'
        }

        metrics = {
            'app.cpuLoad': process_cpu_load if process_cpu_load is not None else -1,
            'sys.cpuLoad': system_cpu_load if system_cpu_load is not None else -1
        }
        cpu_metric_data.update(self.metric_data)
        cpu_metric_data['metrics'] = metrics
        cpu_metric_report = {
            'data': cpu_metric_data,
            'type': 'Metric',
            'apiKey': reporter.api_key,
            'dataModelVersion': constants.DATA_FORMAT_VERSION
        }
        reporter.add_report(cpu_metric_report)



import thundra.utils as utils
from thundra.constants import (
    APPLICATION_ID_PROP_NAME,
    APPLICATION_DOMAIN_NAME_PROP_NAME,
    APPLICATION_CLASS_NAME_PROP_NAME,
    APPLICATION_NAME_PROP_NAME,
    APPLICATION_VERSION_PROP_NAME,
    APPLICATION_STAGE_PROP_NAME,
    APPLICATION_TAG_PROP_NAME_PREFIX,
)



_application_tags = {}
application_domain_name = utils.get_configuration(APPLICATION_DOMAIN_NAME_PROP_NAME)
application_class_name = utils.get_configuration(APPLICATION_CLASS_NAME_PROP_NAME)
application_name = utils.get_configuration(APPLICATION_NAME_PROP_NAME)
application_version = utils.get_configuration(APPLICATION_VERSION_PROP_NAME)
application_id = utils.get_configuration(APPLICATION_ID_PROP_NAME)
application_stage = utils.get_configuration(APPLICATION_STAGE_PROP_NAME)

def get_application_tags():
    return _application_tags.copy()

def clear_application_tags():
    _application_tags.clear()

def parse_application_tags():
    prefix_length = len(APPLICATION_TAG_PROP_NAME_PREFIX)
    for env_key in utils.get_all_env_variables():
        if env_key.startswith(APPLICATION_TAG_PROP_NAME_PREFIX):
            app_tag_key = env_key[prefix_length:]
            env_val = utils.get_configuration(env_key)
            try:
                app_tag_val = utils.str2bool(env_val)
            except ValueError:
                try:
                    app_tag_val = int(env_val)
                except ValueError:
                    try:
                        app_tag_val = float(env_val)
                    except ValueError:
                        app_tag_val = env_val.strip('"')
            
            _application_tags[app_tag_key] = app_tag_val
    return _application_tags.copy()


from __future__ import absolute_import
import thundra.constants as Constants
from thundra.integrations.base_integration import BaseIntegration


class RedisIntegration(BaseIntegration):
    CLASS_TYPE = 'redis'

    def __init__(self):
        pass

    def get_operation_name(self):
        return 'redis_call'

    def getCommandType(self, string):
        if string in Constants.RedisCommandTypes:
            return Constants.RedisCommandTypes[string]
        return 'READ'

    # pylint: disable=W0613
    def inject_span_info(self, scope, wrapped, instance, args, kwargs, response, exception):
        connection = str(instance.connection_pool).split(',')
        host = connection[0][connection[0].find('host=')+5:]
        port = connection[1][connection[1].find('port=')+5:]
        command_type = wrapped.__name__.upper()

        scope.span.domain_name = Constants.DomainNames['CACHE']
        scope.span.class_name = Constants.ClassNames['REDIS']
        scope.span.operation_name = 'REDIS | ' + self.getCommandType(command_type)
        scope.span.operationName = host

        ## ADDING TAGS ##

        tags = {
            Constants.SpanTags['SPAN_TYPE']: Constants.SpanTypes['REDIS'],
            Constants.SpanTags['OPERATION_TYPE']: self.getCommandType(command_type),
            Constants.DBTags['DB_INSTANCE']: host + ':' + port,
            Constants.DBTags['DB_STATEMENT']: {'command': command_type.lower(), 'args': [str(arg) for arg in args]},
            Constants.DBTags['DB_STATEMENT_TYPE']: self.getCommandType(command_type),
            Constants.DBTags['DB_TYPE']: 'redis',
            Constants.RedisTags['REDIS_HOST']: host + ':' + port,
            Constants.RedisTags['REDIS_COMMAND_TYPE']: command_type,
            Constants.RedisTags['REDIS_COMMAND_ARGS']: [str(arg) for arg in args],
            Constants.RedisTags['REDIS_COMMAND']: self.getCommandType(command_type),
        }

        ## FINISHED ADDING TAGS ##
        scope.span.tags = tags


from thundra import utils, constants
import sys
import thundra.application_support as application_support
import wrapt
from thundra.plugins.log.thundra_log_handler import logs
from thundra.plugins.log.thundra_log_handler import ThundraLogHandler
from thundra.opentracing.tracer import ThundraTracer
import logging


class LogPlugin:

    def __init__(self):
        self.hooks = {
            'before:invocation': self.before_invocation,
            'after:invocation': self.after_invocation
        }
        self.log_data = {}
        self.tracer = ThundraTracer.get_instance()
        disable_prints_to_logs_by_env = utils.get_configuration(constants.THUNDRA_LAMBDA_LOG_CONSOLE_PRINT_DISABLE)
        if not utils.should_disable(disable_prints_to_logs_by_env):
            self.logger = logging.getLogger('STDOUT')
            self.handler = ThundraLogHandler()
            self.logger.addHandler(self.handler)
            self.logger.setLevel(logging.INFO)
            self.handler.setLevel(logging.INFO)
            self.logger.propagate = False
            wrapt.wrap_function_wrapper(
                'builtins',
                'print',
                self._wrapper
            )

    def _wrapper(self, wrapped, instance, args, kwargs):
        try:
            wrapped(*args, **kwargs)
            self.logger.info(str(args[0]))
        except:
            pass

    def before_invocation(self, plugin_context):
        logs.clear()

        context = plugin_context['context']
        function_name = getattr(context, constants.CONTEXT_FUNCTION_NAME, None)

        self.log_data = {
            'type': "Log",
            'agentVersion': constants.THUNDRA_AGENT_VERSION,
            'dataModelVersion': constants.DATA_FORMAT_VERSION,
            'applicationId': utils.get_application_id(context),
            'applicationDomainName': constants.AWS_LAMBDA_APPLICATION_DOMAIN_NAME,
            'applicationClassName': constants.AWS_LAMBDA_APPLICATION_CLASS_NAME,
            'applicationName': function_name,
            'applicationVersion': getattr(context, constants.CONTEXT_FUNCTION_VERSION, None),
            'applicationStage': utils.get_configuration(constants.THUNDRA_APPLICATION_STAGE, ''),
            'applicationRuntime': 'python',
            'applicationRuntimeVersion': str(sys.version_info[0]),
            'applicationTags': {},

            'traceId': plugin_context.get('trace_id', ""),
            'transactionId': plugin_context.get('transaction_id', context.aws_request_id),
            'tags': {}
        }

    def after_invocation(self, plugin_context):
        context = plugin_context['context']

        #### ADDING TAGS ####
        self.log_data['applicationTags'] = application_support.get_application_tags()
        self.log_data['tags']['aws.region'] = utils.get_aws_region_from_arn(getattr(context, constants.CONTEXT_INVOKED_FUNCTION_ARN, None))
        self.log_data['tags']['aws.lambda.name'] = getattr(context, constants.CONTEXT_FUNCTION_NAME, None)
        self.log_data['tags']['aws.lambda.arn'] = getattr(context, constants.CONTEXT_INVOKED_FUNCTION_ARN, None)
        self.log_data['tags']['aws.lambda.memory.limit'] = getattr(context, constants.CONTEXT_MEMORY_LIMIT_IN_MB, None)
        self.log_data['tags']['aws.lambda.log_group_name'] = getattr(context, constants.CONTEXT_LOG_GROUP_NAME, None)
        self.log_data['tags']['aws.lambda.log_stream_name'] = getattr(context, constants.CONTEXT_LOG_STREAM_NAME, None)

        if 'error' in plugin_context:
            error = plugin_context['error']
            error_type = type(error)
            # Adding tags
            self.log_data['tags']['error'] = True
            self.log_data['tags']['error.kind'] = error_type.__name__
            self.log_data['tags']['error.message'] = str(error)
            if hasattr(error, 'code'):
                self.log_data['tags']['error.code'] = error.code
            if hasattr(error, 'object'):
                self.log_data['tags']['error.object'] = error.object
            if hasattr(error, 'stack'):
                self.log_data['tags']['error.stack'] = error.stack

        reporter = plugin_context['reporter']
        for log in logs:
            log.update(self.log_data)
            log_report = {
                'data': log,
                'type': 'Log',
                'apiKey': reporter.api_key,
                'dataModelVersion': constants.DATA_FORMAT_VERSION
            }
            reporter.add_report(log_report)

        logs.clear()



from threading import Lock


class RecordEvents:
    START_SPAN = 'start_span'
    FINISH_SPAN = 'finish_span'


class ThundraRecorder:

    def __init__(self):
        self._lock = Lock()
        self._active_span_stack = []
        self._finished_span_stack = []
        self.listeners = []

    @property
    def active_span_stack(self):
        return self._active_span_stack

    @property
    def finished_span_stack(self):
        return self._finished_span_stack

    def get_active_span(self):
        if self._active_span_stack is not None and len(self._active_span_stack) > 0:
            return self._active_span_stack[-1]
        if self._finished_span_stack is not None and len(self._finished_span_stack) > 0:
            return self.finished_span_stack[0]
        return None

    def get_root_span(self):
        if self._active_span_stack is not None and len(self._active_span_stack) > 0:
            return self._active_span_stack[0]
        return None

    def record(self, event, span):
        with self._lock:
            if event == RecordEvents.START_SPAN:
                self._record_start_span(span)
                for listener in self.listeners:
                    listener.on_span_started(span)
            elif event == RecordEvents.FINISH_SPAN:
                self._record_finish_span()
                for listener in self.listeners:
                    listener.on_span_finished(span)

    def clear(self):
        self.listeners.clear()
        self._finished_span_stack.clear()

    def _record_start_span(self, span):
        self._active_span_stack.append(span)

    def _record_finish_span(self):
        if len(self._active_span_stack) > 0:
            self._finished_span_stack.append(self.active_span_stack[-1])
            self._active_span_stack.pop()

    def add_listener(self, listener):
        self.listeners.append(listener)



import logging
from thundra import constants
import thundra.utils as utils

loggers = {}


def get_logger(name):
    global loggers
    if loggers.get(name):
        return loggers.get(name)
    else:
        FORMAT = "%(asctime)s  - %(levelname)s - %(name)s - %(message)s"
        if name is None:
            logger = logging.getLogger(__name__)
        else:
            logger = logging.getLogger(name)
        logger.setLevel(logging.DEBUG)
        console_handler = logging.StreamHandler()
        console_handler.setLevel(logging.DEBUG)
        ch_format = logging.Formatter(FORMAT)
        console_handler.setFormatter(ch_format)
        logger.addHandler(console_handler)
        loggers[name] = logger
        return logger


def log_to_console(message,  handler):
    logger = get_logger(handler)
    logging.getLogger().handlers = []
    logger.debug(message)


def debug_logger(msg, handler=None):
    if utils.get_configuration(constants.THUNDRA_LAMBDA_DEBUG_ENABLE, 'false') == 'true':
        if hasattr(msg, '__dict__'):
            log_to_console(msg, handler)
            display = vars(msg)
            log_to_console(display, handler)
            for key, value in display.items():
                debug_logger_helper(getattr(msg, key), handler)
        else:
            log_to_console(msg, handler)


def debug_logger_helper(msg, handler):
    if hasattr(msg, '__dict__'):
        log_to_console(msg, handler)
        display = vars(msg)
        log_to_console(display,  handler)
        for key, value in display.items():
            debug_logger_helper(getattr(msg, key), handler)




class RdbBaseIntegration():
    _OPERATION_TO_TABLE_NAME_KEYWORD = {
        'select': 'from',
        'insert': 'into',
        'update': 'update',
        'delete': 'from',
        'create': 'table'
    }

    _OPERATION_TO_TYPE = {
        'select': 'READ',
        'insert': 'WRITE',
        'update': 'WRITE',
        'delete': 'DELETE'
    }

    @staticmethod
    def _extract_table_name(query, operation):
        if operation in RdbBaseIntegration._OPERATION_TO_TABLE_NAME_KEYWORD:
            keyword =  RdbBaseIntegration._OPERATION_TO_TABLE_NAME_KEYWORD[operation]
            query_words = query.lower().split()
            if keyword in query_words:
                return query.split()[query_words.index(keyword) + 1]
        return ''

    def set_exception(self, exception, traceback_data, span):
        span.set_tag('error.stack', traceback_data)
        span.set_error_to_tag(exception)


"""
botocore patcher module.
"""

from __future__ import absolute_import
import wrapt
from thundra import utils
from thundra import constants
from thundra.integrations.base_integration import BaseIntegration
import thundra.integrations.botocore
from thundra.integrations.modules.requests import _wrapper as request_wrapper

INTEGRATIONS = {
    class_obj.CLASS_TYPE: class_obj()
    for class_obj in BaseIntegration.__subclasses__()
}


def _wrapper(wrapped, instance, args, kwargs):
    integration_name = instance.__class__.__name__.lower()
    try:
        response = INTEGRATIONS[integration_name].create_span(
            wrapped,
            instance,
            args,
            kwargs
        )
        return response
    except:
        pass
    return wrapped(*args, **kwargs)


def patch():
    disable_aws_integration_by_env = utils.get_configuration(constants.THUNDRA_DISABLE_AWS_INTEGRATION)
    disable_http_integration_by_env = utils.get_configuration(constants.THUNDRA_DISABLE_HTTP_INTEGRATION)
    if not utils.should_disable(disable_aws_integration_by_env):
        wrapt.wrap_function_wrapper(
            'botocore.client',
            'BaseClient._make_api_call',
            _wrapper
        )
    if not utils.should_disable(disable_http_integration_by_env):
        wrapt.wrap_function_wrapper(
            'botocore.vendored.requests',
            'Session.send',
            request_wrapper
        )



from __future__ import absolute_import
import thundra.constants as Constants
from thundra.integrations.base_integration import BaseIntegration
import traceback

# pylint: disable=W0613


def dummy_func(*args):
    return None


def set_exception(exception, traceback_data, scope):
    span = scope.span
    span.set_tag('error.stack', traceback_data)
    span.set_error_to_tag(exception)


class AWSDynamoDBIntegration(BaseIntegration):
    CLASS_TYPE = 'dynamodb'
    OPERATION = {}

    def __init__(self):
        pass

    def get_operation_name(self):
        return 'dynamodb'

    def getStatementType(self, string):
        if string in Constants.DynamoDBRequestTypes:
            return Constants.DynamoDBRequestTypes[string]
        return 'READ'

    def inject_span_info(self, scope, wrapped, instance, args, kwargs, response,
                 exception):

        self.OPERATION.update(
            {'PutItem': self.process_put_item_op,
             'UpdateItem': self.process_update_item_op,
             'GetItem': self.process_get_item_op,
             'DeleteItem': self.process_delete_item_op,
             'BatchWriteItem': self.process_batch_write_op,
             }
        )

        operation_name, request_data = args
        self.request_data = request_data
        self.response = response
        self.endpoint = instance._endpoint.host.split('/')[-1]

        scope.span.domain_name = Constants.DomainNames['DB']
        scope.span.class_name = Constants.ClassNames['DYNAMODB']
        scope.span.operation_name = 'dynamodb: ' + str(self.request_data['TableName'])

        ## ADDING TAGS ##

        tags = {
            Constants.SpanTags['SPAN_TYPE']: Constants.SpanTypes['AWS_DYNAMO'],
            Constants.SpanTags['OPERATION_TYPE']: self.getStatementType(operation_name),
            Constants.DBTags['DB_INSTANCE']: self.endpoint,
            Constants.DBTags['DB_TYPE']: Constants.DBTypes['DYNAMODB'],
            Constants.AwsDynamoTags['TABLE_NAME']: str(self.request_data['TableName']),
            Constants.DBTags['DB_STATEMENT_TYPE']: self.getStatementType(operation_name),
            Constants.AwsSDKTags['REQUEST_NAME']: operation_name,
        }
        scope.span.tags = tags
        ## FINISHED ADDING TAGS ##
        self.OPERATION.get(operation_name, dummy_func)(scope)

        if exception is not None:
            set_exception(exception, traceback.format_exc(), scope)

    def process_get_item_op(self, scope):
        if 'Key' in self.request_data:
            scope.span.set_tag(Constants.DBTags['DB_STATEMENT'], self.request_data['Key'])

    def process_put_item_op(self, scope):
        if 'Item' in self.request_data:
            scope.span.set_tag(Constants.DBTags['DB_STATEMENT'], self.request_data['Item'])

    def process_delete_item_op(self, scope):
        if 'Key' in self.request_data:
            scope.span.set_tag(Constants.DBTags['DB_STATEMENT'], self.request_data['Key'])

    def process_update_item_op(self, scope):
        if 'Key' in self.request_data:
            scope.span.set_tag(Constants.DBTags['DB_STATEMENT'], self.request_data['Key'])

    def process_batch_write_op(self, scope):
            table_name = list(self.request_data['RequestItems'].keys())[0]
            items = []
            for item in self.request_data['RequestItems'][table_name]:
                items.append(item)

            scope.span.set_tag(Constants.DBTags['DB_STATEMENT'], items)


class AWSSQSIntegration(BaseIntegration):
    CLASS_TYPE = 'sqs'

    def __init__(self):
        pass

    def get_operation_name(self):
        return 'sqs'

    def getRequestType(self, string):
        if string in Constants.SQSRequestTypes:
            return Constants.SQSRequestTypes[string]
        return 'READ'

    def getQueueName(self, data):
        if 'QueueUrl' in data:
            return data['QueueUrl'].split('/')[-1]
        elif 'QueueName' in data:
            return data['QueueName']

    def inject_span_info(self, scope, wrapped, instance, args, kwargs, response,
                 exception):

        operation_name, request_data = args
        self.request_data = request_data
        self.queueName = str(self.getQueueName(self.request_data))
        self.response = response

        scope.span.domain_name = Constants.DomainNames['MESSAGING']
        scope.span.class_name = Constants.ClassNames['SQS']
        scope.span.operation_name = self.getRequestType(operation_name)

        ## ADDING TAGS ##

        tags = {
            Constants.SpanTags['SPAN_TYPE']: Constants.SpanTypes['AWS_SQS'],
            Constants.AwsSQSTags['QUEUE_NAME']: self.queueName,
            Constants.SpanTags['OPERATION_TYPE']: self.getRequestType(operation_name),
            Constants.AwsSDKTags['REQUEST_NAME']: operation_name,
        }
        scope.span.tags = tags

        ## FINISHED ADDING TAGS ##

        if exception is not None:
            set_exception(exception, traceback.format_exc(), scope)


class AWSSNSIntegration(BaseIntegration):
    CLASS_TYPE = 'sns'

    def __init__(self):
        pass

    def get_operation_name(self):
        return 'sns'

    def getRequestType(self, string):
        if string in Constants.SNSRequestTypes:
            return Constants.SNSRequestTypes[string]
        return Constants.AWS_SERVICE_REQUEST

    def inject_span_info(self, scope, wrapped, instance, args, kwargs, response,
                 exception):

        operation_name, request_data = args
        self.request_data = request_data
        self.response = response

        scope.span.domain_name = Constants.DomainNames['MESSAGING']
        scope.span.class_name = Constants.ClassNames['SNS']
        scope.span.operation_name = self.getRequestType(operation_name)

        if operation_name == 'CreateTopic':
            self.topicName = request_data.get('Name', 'N/A')
        else:
            arn = request_data.get(
                        'TopicArn',
                        request_data.get('TargetArn', 'N/A')
                    )
            self.topicName = arn.split(':')[-1]

        ### ADDING TAGS ###
        tags = {
            Constants.AwsSDKTags['REQUEST_NAME']: operation_name,
            Constants.SpanTags['OPERATION_TYPE']: self.getRequestType(operation_name),
            Constants.AwsSNSTags['TOPIC_NAME']: self.topicName
        }
        ### FINISHED ADDING TAGS ###
        scope.span.tags = tags

        if exception is not None:
            set_exception(exception, traceback.format_exc(), scope)


class AWSKinesisIntegration(BaseIntegration):
    CLASS_TYPE = 'kinesis'

    def __init__(self):
        pass

    def get_operation_name(self):
        return 'kinesis'

    def getRequestType(self, string):
        if string in Constants.KinesisRequestTypes:
            return Constants.KinesisRequestTypes[string]
        return Constants.AWS_SERVICE_REQUEST

    def inject_span_info(self, scope, wrapped, instance, args, kwargs, response,
                 exception):

        operation_name, request_data = args
        self.streamName = request_data['StreamName']

        scope.span.domain_name = Constants.DomainNames['STREAM']
        scope.span.class_name = Constants.ClassNames['KINESIS']
        scope.span.operation_name = self.getRequestType(operation_name)


        ### ADDING TAGS ###
        tags = {
            Constants.AwsSDKTags['REQUEST_NAME']: operation_name,
            Constants.SpanTags['OPERATION_TYPE']: self.getRequestType(operation_name),
            Constants.AwsKinesisTags['STREAM_NAME']: self.streamName
        }
        ### FINISHED ADDING TAGS ###
        scope.span.tags = tags

        if exception is not None:
            set_exception(exception, traceback.format_exc(), scope)


class AWSFirehoseIntegration(BaseIntegration):
    CLASS_TYPE = 'firehose'

    def __init__(self):
        pass

    def get_operation_name(self):
        return 'firehose'

    def getRequestType(self, string):
        if string in Constants.FirehoseRequestTypes:
            return Constants.FirehoseRequestTypes[string]
        return Constants.AWS_SERVICE_REQUEST

    def inject_span_info(self, scope, wrapped, instance, args, kwargs, response,
                 exception):

        operation_name, request_data = args
        self.deliveryStreamName = request_data['DeliveryStreamName']

        scope.span.domain_name = Constants.DomainNames['STREAM']
        scope.span.class_name = Constants.ClassNames['FIREHOSE']
        scope.span.operation_name = self.getRequestType(operation_name)

        ### ADDING TAGS ###
        tags = {
            Constants.AwsSDKTags['REQUEST_NAME']: operation_name,
            Constants.SpanTags['OPERATION_TYPE']: self.getRequestType(operation_name),
            Constants.AwsFirehoseTags['STREAM_NAME']: self.deliveryStreamName
        }
        ## FINISHED ADDING TAGS ###
        scope.span.tags = tags

        if exception is not None:
            set_exception(exception, traceback.format_exc(), scope)


class AWSS3Integration(BaseIntegration):
    CLASS_TYPE = 's3'

    def __init__(self):
        pass

    def get_operation_name(self):
        return 's3'

    def getRequestType(self, string):
        if string in Constants.S3RequestTypes:
            return Constants.S3RequestTypes[string]
        return Constants.AWS_SERVICE_REQUEST

    def inject_span_info(self, scope, wrapped, instance, args, kwargs, response,
                 exception):

        operation_name, request_data = args
        self.bucket = request_data['Bucket']

        scope.span.domain_name = Constants.DomainNames['STORAGE']
        scope.span.class_name = Constants.ClassNames['S3']
        scope.span.operation_name = self.getRequestType(operation_name)

        if "Key" in request_data:
            self.objectName = request_data["Key"]

        ### ADDING TAGS ###
        tags = {
            Constants.AwsSDKTags['REQUEST_NAME']: operation_name,
            Constants.SpanTags['OPERATION_TYPE']: self.getRequestType(operation_name),
            Constants.AwsS3Tags['BUCKET_NAME']: self.bucket,
            Constants.AwsS3Tags['OBJECT_NAME']: self.objectName
        }
        ## FINISHED ADDING TAGS ###
        scope.span.tags = tags

        if exception is not None:
            set_exception(exception, traceback.format_exc(), scope)


class AWSLambdaIntegration(BaseIntegration):
    CLASS_TYPE = 'lambda'

    def __init__(self):
        pass

    def get_operation_name(self):
        return 'lambda'

    def getRequestType(self, string):
        if string in Constants.LambdaRequestType:
            return Constants.LambdaRequestType[string]
        return Constants.AWS_SERVICE_REQUEST

    def inject_span_info(self, scope, wrapped, instance, args, kwargs, response,
                 exception):

        operation_name, request_data = args
        self.lambdaFunction = request_data.get('FunctionName', '')
        scope.span.domain_name = Constants.DomainNames['API']
        scope.span.class_name = Constants.ClassNames['LAMBDA']
        scope.span.operation_name = self.getRequestType(operation_name)

        ### ADDING TAGS ###
        tags = {
            Constants.AwsSDKTags['REQUEST_NAME']: operation_name,
            Constants.SpanTags['OPERATION_TYPE']: self.getRequestType(operation_name),
            Constants.AwsLambdaTags['FUNCTION_NAME']: self.lambdaFunction,
        }
        if 'Payload' in request_data:
            tags[Constants.AwsLambdaTags['INVOCATION_PAYLOAD']] = request_data['Payload']

        if 'Qualifier' in request_data:
            tags[Constants.AwsLambdaTags['FUNCTION_QUALIFIER']] = request_data['Qualifier']

        if 'InvocationType' in request_data:
            tags[Constants.AwsLambdaTags['INVOCATION_TYPE']] = request_data['InvocationType']
        ## FINISHED ADDING TAGS ###

        scope.span.tags = tags

        if exception is not None:
            set_exception(exception, traceback.format_exc(), scope)


class AWSXrayIntegration(BaseIntegration):
    CLASS_TYPE = 'xray'

    def __init__(self):
        pass

    def get_operation_name(self):
        return 'xray'

    # def getRequestType(self, string):
    #     if string in Constants.LambdaRequestType:
    #         return Constants.LambdaRequestType[string]
    #     return Constants.AWS_SERVICE_REQUEST

    def inject_span_info(self, scope, wrapped, instance, args, kwargs, response,
                 exception):

        operation_name, request_data = args
        # self.lambdaFunction = request_data.get('FunctionName', '')
        # scope.span.domain_name = Constants.DomainNames['API']
        scope.span.class_name = 'XRAY'
        # scope.span.operation_name = 'lambda: ' + self.lambdaFunction

        ### ADDING TAGS ###
        tags = {
            # Constants.AwsSDKTags['REQUEST_NAME']: operation_name,
            # Constants.SpanTags['OPERATION_TYPE']: self.getRequestType(operation_name),
            # Constants.AwsLambdaTags['FUNCTION_NAME']: self.lambdaFunction,
            "XRAY": 'Under_Development'
        }
        if 'Payload' in request_data:
            tags[Constants.AwsLambdaTags['INVOCATION_PAYLOAD']] = request_data['Payload']

        if 'Qualifier' in request_data:
            tags[Constants.AwsLambdaTags['FUNCTION_QUALIFIER']] = request_data['Qualifier']

        if 'InvocationType' in request_data:
            tags[Constants.AwsLambdaTags['INVOCATION_TYPE']] = request_data['InvocationType']
        ## FINISHED ADDING TAGS ###

        scope.span.tags = tags

        if exception is not None:
            set_exception(exception, traceback.format_exc(), scope)



DATE_FORMAT = "%Y-%m-%d %H:%M:%S.%f %z"

HOST = "https://api.thundra.io"
PATH = "/v1/monitoring-data"

REQUEST_COUNT = 0

THUNDRA_APIKEY = 'thundra_apiKey'
THUNDRA_DISABLE = 'thundra_agent_lambda_disable'
THUNDRA_APPLICATION_DOMAIN_NAME = 'thundra_agent_lambda_application_domainName'
THUNDRA_APPLICATION_CLASS_NAME = 'thundra_agent_lambda_application_className'
THUNDRA_APPLICATION_STAGE = 'thundra_agent_lambda_application_stage'
THUNDRA_APPLICATION_VERSION = 'thundra_agent_lambda_application_version'
THUNDRA_AGENT_VERSION = '2.0.0'

THUNDRA_LAMBDA_TIMEOUT_MARGIN = 'thundra_agent_lambda_timeout_margin'
THUNDRA_LAMBDA_REPORT_REST_BASEURL = 'thundra_agent_lambda_report_rest_baseUrl'
THUNDRA_LAMBDA_REPORT_CLOUDWATCH_ENABLE = 'thundra_agent_lambda_report_cloudwatch_enable'
THUNDRA_DISABLE_TRACE = 'thundra_agent_lambda_trace_disable'
THUNDRA_DISABLE_METRIC = 'thundra_agent_lambda_metric_disable'
THUNDRA_DISABLE_LOG = 'thundra_agent_lambda_log_disable'
THUNDRA_DISABLE_AWS_INTEGRATION = 'thundra_agent_lambda_trace_integration_aws_disable'
THUNDRA_DISABLE_REDIS_INTEGRATION = 'thundra_agent_lambda_trace_integration_redis_disable'

THUNDRA_LAMBDA_TRACE_REQUEST_SKIP = 'thundra_agent_lambda_trace_request_skip'
THUNDRA_LAMBDA_TRACE_RESPONSE_SKIP = 'thundra_agent_lambda_trace_response_skip'
THUNDRA_LAMBDA_TRACE_INSTRUMENT_DISABLE = 'thundra_agent_lambda_trace_instrument_disable'
THUNDRA_LAMBDA_TRACE_INSTRUMENT_CONFIG = 'thundra_agent_lambda_trace_instrument_traceableConfig'
THUNDRA_LAMBDA_TRACE_ENABLE_XRAY = 'thundra_agent_lambda_trace_enable_xray'

THUNDRA_DISABLE_HTTP_INTEGRATION = 'thundra_agent_lambda_trace_integrations_http_disable'
THUNDRA_DISABLE_RDB_INTEGRATION = 'thundra_agent_lambda_trace_integrations_rdb_disable'

AWS_LAMBDA_APPLICATION_ID = 'AWS_LAMBDA_APPLICATION_ID'
AWS_LAMBDA_LOG_STREAM_NAME = 'AWS_LAMBDA_LOG_STREAM_NAME'
AWS_LAMBDA_FUNCTION_VERSION = 'AWS_LAMBDA_FUNCTION_VERSION'
AWS_REGION = 'AWS_REGION'
AWS_LAMBDA_FUNCTION_MEMORY_SIZE = "AWS_LAMBDA_FUNCTION_MEMORY_SIZE"
AWS_LAMBDA_APPLICATION_DOMAIN_NAME = 'API'
AWS_LAMBDA_APPLICATION_CLASS_NAME = 'AWS-Lambda'

CONTEXT_FUNCTION_NAME = 'function_name'
CONTEXT_FUNCTION_VERSION = 'function_version'
CONTEXT_FUNCTION_PLATFORM = 'AWS-Lambda'
CONTEXT_MEMORY_LIMIT_IN_MB = 'memory_limit_in_mb'
CONTEXT_INVOKED_FUNCTION_ARN = 'invoked_function_arn'
CONTEXT_AWS_REQUEST_ID = 'aws_request_id'
CONTEXT_LOG_GROUP_NAME = 'log_group_name'
CONTEXT_LOG_STREAM_NAME = 'log_stream_name'

APPLICATION_ID_PROP_NAME = 'thundra_agent_lambda_application_id'
APPLICATION_DOMAIN_NAME_PROP_NAME = 'thundra_agent_lambda_application_domainName'
APPLICATION_CLASS_NAME_PROP_NAME = 'thundra_agent_lambda_application_className'
APPLICATION_NAME_PROP_NAME = 'thundra_agent_lambda_application_name'
APPLICATION_VERSION_PROP_NAME = 'thundra_agent_lambda_application_version'
APPLICATION_STAGE_PROP_NAME = 'thundra_agent_lambda_application_stage'
APPLICATION_TAG_PROP_NAME_PREFIX = 'thundra_agent_lambda_application_tag_'

TRACE_ARGS = 'trace_args'
TRACE_RETURN_VALUE = 'trace_return_value'
TRACE_ERROR = 'trace_error'

LIST_SEPARATOR = ','

DEFAULT_LAMBDA_TIMEOUT_MARGIN = 200
DATA_FORMAT_VERSION = '2.0'

THUNDRA_LAMBDA_DEBUG_ENABLE = 'thundra_agent_lambda_debug_enable'
THUNDRA_LAMBDA_LOG_CONSOLE_PRINT_DISABLE = 'thundra_agent_lambda_log_console_disable'

#### INTEGRATIONS ####

AWS_SERVICE_REQUEST = 'AWSServiceRequest'

DomainNames = {
    'AWS': 'AWS',
    'DB': 'DB',
    'MESSAGING': 'Messaging',
    'STREAM': 'Stream',
    'STORAGE': 'Storage',
    'API': 'API',
    'CACHE': 'Cache',
}

ClassNames = {
    'AWSSERVICE': 'AWSService',
    'DYNAMODB': 'AWS-DynamoDB',
    'SQS': 'AWS-SQS',
    'SNS': 'AWS-SNS',
    'KINESIS': 'AWS-Kinesis',
    'FIREHOSE': 'AWS-Firehose',
    'S3': 'AWS-S3',
    'LAMBDA': 'AWS-Lambda',
    'RDB': 'RDB',
    'REDIS': 'Redis',
    'HTTP': 'HTTP',
}

DBTags = {
    'DB_STATEMENT': 'db.statement',
    'DB_STATEMENT_TYPE': 'db.statement.type',
    'DB_INSTANCE': 'db.instance',
    'DB_TYPE': 'db.type',
    'DB_HOST': 'db.host',
    'DB_PORT': 'db.port',
    'DB_USER': 'db.user',
}

AwsDynamoTags = {
    'TABLE_NAME': 'aws.dynamodb.table.name',
    'REQUEST_THROTTLED': 'aws.dynamodb.request.throttled',
}

DBTypes = {
    'DYNAMODB': 'aws-dynamodb',
    'REDIS': 'redis',
}

SpanTags = {
    'SPAN_TYPE': 'span.type',
    'OPERATION_TYPE': 'operation.type',
    'DB_INSTANCE': 'db.instance',
    'DB_TYPE': 'db.type',
    'DB_TABLE_NAME': 'db.table.name',
    'DB_HOST': 'db.host',
    'TRIGGER_DOMAIN_NAME': 'trigger.domainName',
    'TRIGGER_CLASS_NAME': 'trigger.className',
    'DB_STATEMENT': 'db.statement',
    'DB_STATEMENT_TYPE': 'db.statement.type:',
}

SpanTypes = {
    'REDIS': 'Redis',
    'RDB': 'RDB',
    'HTTP': 'HTTP',
    'AWS_DYNAMO': 'AWS-DynamoDB',
    'AWS_SQS': 'AWS-SQS',
    'AWS_SNS': 'AWS-SNS',
    'AWS_KINESIS': 'AWS-Kinesis',
    'AWS_FIREHOSE': 'AWS-Firehose',
    'AWS_S3': 'AWS-S3',
    'AWS_LAMBDA': 'AWS-Lambda',
}

DynamoDBRequestTypes = {
    'BatchGetItem': 'READ',
    'BatchWriteItem': 'WRITE',
    'CreateTable': 'WRITE',
    'CreateGlobalTable': 'WRITE',
    'DeleteItem': 'DELETE',
    'DeleteTable': 'DELETE',
    'GetItem': 'READ',
    'PutItem': 'WRITE',
    'Query': 'READ',
    'Scan': 'READ',
    'UpdateItem': 'WRITE',

}

AwsSDKTags = {
    'SERVICE_NAME': 'aws.service.name',
    'REQUEST_NAME': 'aws.request.name',
    'HOST': 'host',
}

SQSRequestTypes = {
    'ReceiveMessage': 'READ',
    'SendMessage': 'WRITE',
    'SendMessageBatch': 'WRITE',
    'DeleteMessage': 'DELETE',
    'DeleteMessageBatch': 'DELETE',
}

AwsSQSTags = {
    'QUEUE_NAME': 'aws.sqs.queue.name',
}

SNSRequestTypes = {
    'Publish': 'WRITE',
}

AwsSNSTags = {
    'TOPIC_NAME': 'aws.sns.topic.name',
}

AwsKinesisTags = {
    'STREAM_NAME': 'aws.kinesis.stream.name',
}

KinesisRequestTypes = {
    'GetRecords': 'READ',
    'PutRecords': 'WRITE',
    'PutRecord': 'WRITE',
}

AwsFirehoseTags = {
    'STREAM_NAME': 'aws.firehose.stream.name',
}

FirehoseRequestTypes = {
    'PutRecordBatch': 'WRITE',
    'PutRecord': 'WRITE',
}

S3RequestTypes = {
    'DeleteBucket': 'DELETE',
    'CreateBucket': 'WRITE',
    'copyObject': 'WRITE',
    'DeleteObject': 'DELETE',
    'deleteObjects': 'DELETE',
    'GetObject': 'READ',
    'GetObjectAcl': 'READ',
    'ListBucket': 'READ',
    'PutObject': 'WRITE',
    'PutObjectAcl': 'WRITE',
}

AwsS3Tags = {
    'BUCKET_NAME': 'aws.s3.bucket.name',
    'OBJECT_NAME': 'aws.s3.object.name',
}

AwsLambdaTags = {
    'FUNCTION_NAME': 'aws.lambda.name',
    'FUNCTION_QUALIFIER': 'aws.lambda.qualifier',
    'INVOCATION_TYPE': 'aws.lambda.invocation.type',
    'INVOCATION_PAYLOAD': 'aws.lambda.invocation.payload',
}

LambdaRequestType = {
    'InvokeAsync': 'CALL',
    'Invoke': 'CALL',
}

HttpTags = {
    'HTTP_METHOD': 'http.method',
    'HTTP_URL': 'http.url',
    'HTTP_PATH': 'http.path',
    'HTTP_HOST': 'http.host',
    'HTTP_STATUS': 'http.status_code',
    'QUERY_PARAMS': 'http.query_params',
}

RedisCommandTypes = {
    'APPEND': 'WRITE',
    'BGREWRITEAOF': 'WRITE',
    'BGSAVE': 'WRITE',
    'BITCOUNT': 'READ',
    'BITFIELD': 'WRITE',
    'BITOP': 'WRITE',
    'BITPOS': 'READ',
    'BLPOP': 'DELETE',
    'BRPOP': 'DELETE',
    'BRPOPLPUSH': 'WRITE',
    'BZPOPMIN': 'DELETE',
    'BZPOPMAX': 'DELETE',
    'DBSIZE': 'READ',
    'DECR': 'WRITE',
    'DECRBY': 'WRITE',
    'DELETE': 'DELETE',
    'EVAL': 'EXECUTE',
    'EVALSHA': 'EXECUTE',
    'EXISTS': 'READ',
    'EXPIRE': 'WRITE',
    'EXPIREAT': 'WRITE',
    'FLUSHALL': 'DELETE',
    'FLUSHDB': 'DELETE',
    'GEOADD': 'WRITE',
    'GEOHASH': 'READ',
    'GEOPOS': 'READ',
    'GEODIST': 'READ',
    'GEORADIUS': 'READ',
    'GEORADIUSBYMEMBER': 'READ',
    'GET': 'READ',
    'GETBIT': 'READ',
    'GETRANGE': 'READ',
    'GETSET': 'WRITE',
    'HDEL': 'DELETE',
    'HEXISTS': 'READ',
    'HGET': 'READ',
    'HGETALL': 'READ',
    'HINCRBY': 'WRITE',
    'HINCRBYFLOAT': 'WRITE',
    'HKEYS': 'READ',
    'HLEN': 'READ',
    'HMGET': 'READ',
    'HMSET': 'WRITE',
    'HSET': 'WRITE',
    'HSETNX': 'WRITE',
    'HSTRLEN': 'READ',
    'HVALS': 'READ',
    'INCR': 'WRITE',
    'INCRBY': 'WRITE',
    'INCRBYFLOAT': 'WRITE',
    'KEYS': 'READ',
    'LINDEX': 'READ',
    'LINSERT': 'WRITE',
    'LLEN': 'READ',
    'LPOP': 'DELETE',
    'LPUSH': 'WRITE',
    'LPUSHX': 'WRITE',
    'LRANGE': 'READ',
    'LREM': 'DELETE',
    'LSET': 'WRITE',
    'LTRIM': 'DELETE',
    'MGET': 'READ',
    'MSET': 'WRITE',
    'MSETNX': 'WRITE',
    'PERSIST': 'WRITE',
    'PEXPIRE': 'WRITE',
    'PEXPIREAT': 'WRITE',
    'PFADD': 'WRITE',
    'PFCOUNT': 'READ',
    'PFMERGE': 'WRITE',
    'PSETEX': 'WRITE',
    'PUBLISH': 'WRITE',
    'RPOP': 'DELETE',
    'RPOPLPUSH': 'WRITE',
    'RPUSH': 'WRITE',
    'RPUSHX': 'WRITE',
    'SADD': 'WRITE',
    'SCARD': 'READ',
    'SDIFFSTORE': 'WRITE',
    'SET': 'WRITE',
    'SETBIT': 'WRITE',
    'SETEX': 'WRITE',
    'SETNX': 'WRITE',
    'SETRANGE': 'WRITE',
    'SINTER': 'READ',
    'SINTERSTORE': 'WRITE',
    'SISMEMBER': 'READ',
    'SMEMBERS': 'READ',
    'SMOVE': 'WRITE',
    'SORT': 'WRITE',
    'SPOP': 'DELETE',
    'SRANDMEMBER': 'READ',
    'SREM': 'DELETE',
    'STRLEN': 'READ',
    'SUNION': 'READ',
    'SUNIONSTORE': 'WRITE',
    'ZADD': 'WRITE',
    'ZCARD': 'READ',
    'ZCOUNT': 'READ',
    'ZINCRBY': 'WRITE',
    'ZINTERSTORE': 'WRITE',
    'ZLEXCOUNT': 'READ',
    'ZPOPMAX': 'DELETE',
    'ZPOPMIN': 'DELETE',
    'ZRANGE': 'READ',
    'ZRANGEBYLEX': 'READ',
    'ZREVRANGEBYLEX': 'READ',
    'ZRANGEBYSCORE': 'READ',
    'ZRANK': 'READ',
    'ZREM': 'DELETE',
    'ZREMRANGEBYLEX': 'DELETE',
    'ZREMRANGEBYRANK': 'DELETE',
    'ZREMRANGEBYSCORE': 'DELETE',
    'ZREVRANGE': 'READ',
    'ZREVRANGEBYSCORE': 'READ',
    'ZREVRANK': 'READ',
    'ZSCORE': 'READ',
    'ZUNIONSTORE': 'WRITE',
    'SCAN': 'READ',
    'SSCAN': 'READ',
    'HSCAN': 'READ',
    'ZSCAN': 'READ',
    'XADD': 'WRITE',
    'XRANGE': 'READ',
    'XREVRANGE': 'READ',
    'XLEN': 'READ',
    'XREAD': 'READ',
    'XREADGROUP': 'READ',
    'XPENDING': 'READ',
}

RedisTags = {
    'REDIS_HOST': 'redis.host',
    'REDIS_PORT': 'redis.port',
    'REDIS_COMMAND': 'redis.command',
    'REDIS_COMMANDS': 'redis.commands',
    'REDIS_COMMAND_TYPE': 'redis.command.type',
    'REDIS_COMMAND_ARGS': 'redis.command.args',
}

AwsXrayConstants = {
    'DEFAULT_OPERATION_NAME': 'AWS X-Ray',
    'XRAY_SUBSEGMENTED_TAG_NAME': 'THUNDRA::XRAY_SUBSEGMENTED',
}


import os
from importlib import import_module
from thundra.thundra_agent import Thundra

thundra = Thundra()

handler_found = False
handler_path = os.environ.get('thundra_agent_lambda_handler', None)
if handler_path is None:
    raise ValueError(
        "No handler specified for \'thundra_agent_lambda_handler\' environment variable"
    )
else:
    handler_found = True
    (module_name, handler_name) = handler_path.rsplit('.', 1)
    user_module = import_module(module_name)
    user_handler = getattr(user_module, handler_name)

@thundra
def wrapper(event, context):
    if handler_found:
        return user_handler(event, context)



from thundra.opentracing.tracer import ThundraTracer
import sys
import thundra.application_support as application_support
from thundra import utils
from thundra import constants

xray_listener_imported = True
try:
    from thundra.listeners.aws_xray_listeners import AWSXrayListener
except ImportError:
    xray_listener_imported = False


class AWSXRayPlugin:

    _data = None

    def __init__(self, listener=None):
        if xray_listener_imported:
            self.hooks = {
                'before:invocation': self.before_invocation,
                'after:invocation': self.after_invocation
            }
            self.tracer = ThundraTracer.get_instance()
            self.xray_data = {}
            if listener is None:
                self.xray_listener = AWSXrayListener()
            else:
                self.xray_listener = listener
            if self.tracer:
                self.tracer.add_span_listener(self.xray_listener)

    def before_invocation(self, plugin_context):
        context = plugin_context['context']
        function_name = getattr(context, constants.CONTEXT_FUNCTION_NAME, None)
        app_tags = application_support.get_application_tags()

        self.xray_data = {
            'applicationId': utils.get_application_id(context),
            'applicationName': function_name,
            'applicationVersion': getattr(context, constants.CONTEXT_FUNCTION_VERSION, None),
            'applicationStage': utils.get_configuration(constants.THUNDRA_APPLICATION_STAGE, ''),
            'applicationRuntime': 'python',
            'applicationRuntimeVersion': str(sys.version_info[0]),
            'applicationTags': app_tags,
        }

        self.send_data(self.xray_data)

    def after_invocation(self, plugin_context):
        pass

    def send_data(self, data):
        if xray_listener_imported:
            self.xray_listener.set_data(data)


import json
import logging

try:
    import requests
except ImportError:
    from botocore.vendored import requests

from thundra import constants
from thundra.utils import get_configuration
import thundra.utils as utils

logger = logging.getLogger(__name__)

class Reporter():

    def __init__(self, api_key, session=None):
        if api_key is not None:
            self.api_key = api_key
        else:
            self.api_key = ''
            logger.error('Please set an API key!')
        self.reports = []

        if not session:
            session = requests.Session()
        self.session = session

    def add_report(self, report):
        if get_configuration(constants.THUNDRA_LAMBDA_REPORT_CLOUDWATCH_ENABLE) == 'true':
            print(json.dumps(report))
        else:
            if isinstance(report, list):
                for data in report:
                    self.reports.append(data)
            else:
                self.reports.append(report)

    def send_report(self):
        headers = {
            'Content-Type': 'application/json',
            'Authorization': 'ApiKey ' + self.api_key
        }
        request_url = constants.HOST + constants.PATH
        base_url = utils.get_configuration(constants.THUNDRA_LAMBDA_REPORT_REST_BASEURL)
        if base_url is not None:
            request_url = base_url + '/monitoring-data'

        response = self.session.post(request_url, headers=headers, data=json.dumps(self.reports))
        self.reports.clear()
        return response



import thundra.constants as constants
import re
from thundra.listeners.thundra_span_listener import ThundraSpanListener
from thundra.opentracing.tracer import ThundraTracer
from aws_xray_sdk.core import xray_recorder

class AWSXrayListener(ThundraSpanListener):

    _data = None
    _tracer = ThundraTracer.get_instance()

    def __init__(self):
        self.xray_data = {}

    def start_subsegment(self, span):
        xray_recorder.begin_subsegment(self.normalize_operation_name(span.operation_name))

    def on_span_started(self, span):
        span.tags[constants.AwsXrayConstants['XRAY_SUBSEGMENTED_TAG_NAME']] = True
        self.start_subsegment(span)

    def end_subsegment(self, span):
        if xray_recorder.current_subsegment():
            xray_recorder.end_subsegment()

    def on_span_finished(self, span):
        self.add_metadata(span)
        self.add_annotation(span)
        self.add_error(span)
        self.end_subsegment(span)

    def put_annotation_if_available(self, key, dictionary, dict_key, document):
        if dict_key in dictionary:
            value = dictionary[dict_key]
            if isinstance(value, str) or isinstance(value, int) or isinstance(value, bool):
                document.put_annotation(self.normalize_annotation_name(key), self.normalize_annotation_value(value))

    def add_annotation(self, span):
        self.xray_data = AWSXrayListener._data
        document = xray_recorder.current_subsegment()
        if document:
            document.put_annotation('traceId', span.context.trace_id)
            document.put_annotation('transactionId', span.context.transaction_id)
            document.put_annotation('spanId', span.context.span_id)
            document.put_annotation('parentSpanId', span.context.parent_span_id)
            span_dict = vars(span)
            self.put_annotation_if_available('domainName', span_dict, 'domain_name', document)
            self.put_annotation_if_available('className', span_dict, 'class_name', document)
            self.put_annotation_if_available('operationName', span_dict, 'operation_name', document)
            self.put_annotation_if_available('startTimeStamp', span_dict, 'start_time', document)
            self.put_annotation_if_available('finishTimeStamp', span_dict, 'finish_time', document)
            duration = int(span.get_duration())
            document.put_annotation('duration', duration)

            if self.xray_data is not None:
                for key in self.xray_data:
                    self.put_annotation_if_available(key, self.xray_data, key, document)

    def add_metadata(self, span):
        document = xray_recorder.current_subsegment()
        if document:
            for key in span.tags:
                document.put_metadata(key, span.tags[key])

    def normalize_operation_name(self, operation_name):
        if operation_name:
            return operation_name[0:200]
        return constants.AwsXrayConstants['DEFAULT_OPERATION_NAME']

    def normalize_annotation_value(self, value):
        if isinstance(value, str):
            value = value[0:1000]
        return value

    def normalize_annotation_name(self, annotation_name):
        annotation_name = re.sub('/\./g', '_', annotation_name)
        annotation_name = re.sub('/[W]+/g', '', annotation_name)
        annotation_name = annotation_name[0:500]
        return annotation_name

    def add_error(self, span):
        if span.get_tag('error'):
            document = xray_recorder.current_subsegment()
            if document:
                document.add_exception(Exception(span.get_tag('error.message')))

    @staticmethod
    def set_data(data):
        AWSXrayListener._data = data



from __future__ import absolute_import
import traceback
import thundra.constants as constants
from urllib.parse import urlparse
from thundra.integrations.base_integration import BaseIntegration


class RequestsIntegration(BaseIntegration):
    CLASS_TYPE = 'http'

    def __init__(self):
        pass

    def get_operation_name(self):
        return 'http_call'
    
    def inject_span_info(self, scope, wrapped, instance, args, kwargs, response, exception):
        prepared_request = args[0]
        method = prepared_request.method
        url = prepared_request.url
        parsed_url = urlparse(url)
        path = parsed_url.path
        query = parsed_url.query
        host = parsed_url.netloc
        span = scope.span
        
        span.operation_name = url
        span.domain_name =  constants.DomainNames['API']
        span.class_name =  constants.ClassNames['HTTP']

        ## ADDING TAGS ##

        tags = {
            constants.SpanTags['SPAN_TYPE']: constants.SpanTypes['HTTP'],
            constants.SpanTags['OPERATION_TYPE']: 'CALL',
            constants.HttpTags['HTTP_METHOD']: method,
            constants.HttpTags['HTTP_URL']: url,
            constants.HttpTags['HTTP_PATH']: path,
            constants.HttpTags['HTTP_HOST']: host,
            constants.HttpTags['QUERY_PARAMS']: query,
        }

        span.tags = tags

        if exception is not None:
            self.set_exception(exception, traceback.format_exc(), span)
        
        if response is not None:
            self.set_response(response, span)

    def set_exception(self, exception, traceback_data, span):
        span.set_tag('error.stack', traceback_data)
        span.set_error_to_tag(exception)
    
    def set_response(self, response, span):
        statusCode = response.status_code
        span.set_tag(constants.HttpTags['HTTP_STATUS'], statusCode)



from __future__ import absolute_import
import traceback
import thundra.constants as constants
from thundra.integrations.base_integration import BaseIntegration
from thundra.integrations.rdb_base import RdbBaseIntegration

try:
    from psycopg2.extensions import parse_dsn

except ImportError:
    def parse_dsn(dsn):
        return dict(
            attribute.split("=") for attribute in dsn.split()
            if "=" in attribute
        )

class PostgreIntegration(BaseIntegration, RdbBaseIntegration):

    def __init__(self):
        pass

    def get_operation_name(self):
        return "postgresql"

    def inject_span_info(self, scope, cursor, connection, _args, _kwargs, response, exception):
        span = scope.span
        span.domain_name = constants.DomainNames['DB']
        span.class_name = constants.ClassNames['RDB']

        dsn = parse_dsn(connection.dsn)
        query = str(cursor.__self__.query)[2:-1].lower()
        operation = query.split()[0].strip("\"").lower()
        tableName = self._extract_table_name(query, operation)

        tags = {
            constants.SpanTags['SPAN_TYPE']: constants.SpanTypes['RDB'],
            constants.SpanTags['OPERATION_TYPE']: PostgreIntegration._OPERATION_TO_TYPE[operation],
            constants.SpanTags['DB_INSTANCE']: dsn['dbname'],
            constants.SpanTags['DB_HOST']: dsn['host'],
            constants.SpanTags['DB_TABLE_NAME']: tableName,
            constants.SpanTags['DB_TYPE']: "postgresql",
            constants.SpanTags['DB_STATEMENT']: query,
            constants.SpanTags['DB_STATEMENT_TYPE']: operation.upper(),
            constants.SpanTags['TRIGGER_DOMAIN_NAME']: "AWS-Lambda",
            constants.SpanTags['TRIGGER_CLASS_NAME']: "API"
        }

        span.tags = tags

        if exception is not None:
            self.set_exception(exception, traceback.format_exc(), span)


import abc
import time
import traceback
from thundra.opentracing.tracer import ThundraTracer


class BaseIntegration(abc.ABC):

    CLASS_TYPE = "base"

    def create_span(self, wrapped, instance, args, kwargs):
        tracer = ThundraTracer.get_instance()
        response = None
        exception = None

        with tracer.start_active_span(operation_name=self.get_operation_name(), finish_on_close=True) as scope:
            try:
                response = wrapped(*args, **kwargs)
                return response
            except Exception as operation_exception:
                exception = operation_exception
                raise
            finally:
                try:
                    self.inject_span_info(scope, wrapped, instance, args, kwargs, response, exception)
                except Exception as instrumentation_exception:
                    error = {
                        'type': str(type(instrumentation_exception)),
                        'message': str(instrumentation_exception),
                        'traceback': traceback.format_exc(),
                        'time': time.time()
                    }
                    scope.span.set_tag('instrumentation_error', error)

    @abc.abstractmethod
    def get_operation_name(self):
        raise Exception("should be implemented")


    @abc.abstractmethod
    def inject_span_info(self, scope, wrapped, instance, args, kwargs, response, exception):
        raise Exception("should be implemented")


import abc


class ThundraSpanListener(abc.ABC):

    @abc.abstractmethod
    def on_span_started(self, operation_name):
        raise Exception("should be implemented")

    @abc.abstractmethod
    def on_span_finished(self, span):
        raise Exception("should be implemented")